# This script uses a training set (object D.train) to generate density functions for two classes: resistant and susceptible
# Samples from these density functions are used to determine best cutoffs for sample means and Bayes factors
# Optimal cutoff values are used to make decisions for samples stored in test set (object D.test)

# Author: Armin O. Schmitt
# September 14, 2021

# File for results
prediction.file <- "PredictionSet.txt"

# File containing training data
# The training set must have three columns
# The first column contains an ID, the second column contains the OD measurement, the third the label: either "resistant" or "susceptible"
# The corresponding column headers must be "ID", "X", and "Label"
training.file <- "TrainingSet.txt"

# File containing training data to be classified
# The data to be classified must contain an ID in the first column
# All other columns contain the OD measurements
# The corresponding column headers must be ID
measurement.file <- "DataSet.txt"


library(ROCit) # For ROC analysis
library(logspline)
options(width = 250)

################################################################################################################################################################################
#
# Reading in training data and data to be classified (new data)
#
################################################################################################################################################################################


cat("Reading in training data...\t")
D.train <- read.table(training.file, header = T)
cat("done\n")

cat("Reading in new data...\t")
D.test <- read.table(measurement.file, header = T)
cat("done\n")


n.train <- nrow(D.train)
n.test <- nrow(D.test)
cat("Number of records for training:\t", n.train, "\n")
cat("Number of records for classification:\t", n.test, "\n")


################################################################################################################################################################################
#
# Some preprocessing and quality control of training data and data to be classified (new data)
#
################################################################################################################################################################################

D.train$Label <- as.factor(D.train$Label) # Make sure "Label" is a factor
labels <- levels(D.train$Label)

n.label <- length(labels)
cat("The training set contains the following labels:\n")
print(labels)

if(n.label != 2) { stop("There must be exactly two labels\n") }

for(i in 1 : n.label)
{
	cat("Label", labels[i], "\t")
	B <- subset(D.train, Label == labels[i])
	group.mean <- mean(B$X)
	group.sd <- sd(B$X)
	cat("mean:\t", group.mean, "sd:\t", group.sd, "\n")
}

x.min <- min(D.train$X); x.max <- max(D.train$X)

cat("The range of data is from", x.min, "to", x.max, "\n")


D.test$ID <- as.factor(D.test$ID) # Make sure ID is treated as factor even if it is a number

# Determine number of columns harbouring numeric values. These are taken as OD measurements
n.col <- ncol(D.test)
measurement.var <- c() # vector for measurements
for(i in 1 : n.col)
{
	var <- names(D.test)[i]
	# cat("column header:\t", var, "\t")
	v <- D.test[, var]
	if(is.numeric(v))  # check each column if values are numeric
	{	# column headers of the numeric columns
		measurement.var <- c(measurement.var, var) 
		# cat("variable is numeric\n")
	}
	# else{ cat("variable is not numeric\n") }
}
	
n.sample <- length(measurement.var)
cat("There are", n.sample, "numeric (measurement) variables\n")


################################################################################################################################################################################
#
# Calculate predictors (sample mean and Bayes factor) for the data to be classified
#
################################################################################################################################################################################


# Density functions for resistant and susceptible training data
set.seed(1)
fit.1 <- logspline(D.train$X[D.train$Label == labels[1]], lbound = x.min, ubound = x.max)
fit.2 <- logspline(D.train$X[D.train$Label == labels[2]], lbound = x.min, ubound = x.max)

#stop()

# generate matrix M that contains only OD measurements
M <- D.test[, names(D.test) %in% measurement.var]
D.test$Mean <- apply(M, MARGIN = 1, FUN = mean) # Mean of the measurements for each sample
BF <- vector(length = n.test) # vector for Bayes factors
# Calculation of Bayes factor for each sample
for(i in 1 : n.test) { BF[i] <- prod(dlogspline(M[i,], fit.1)) / prod(dlogspline(M[i,], fit.2)) }
D.test <- cbind(D.test, BF)


################################################################################################################################################################################
#
# Finding the optimal cutoff values (for sample means and Bayes factors)
#	(i) Simulation of massive drawing of small samples from resistant and susceptible training distributions (object A)
#	(ii) Calcuation of means and Bayes factors of these samples
#	(iii) ROC analysis yield optimal cutoffs
#
################################################################################################################################################################################

##################################################################
# (i) Draw n.rep times small samples (e.g. n.sample = 3) from training distributions
##################################################################

n.rep <- 10000
cat("Drawing", n.rep, "samples of size", n.sample, "...\t")
Label.1 <- rep(labels[1], n.rep)
Mean.1 <- vector(length = n.rep, mode = "numeric")
BF.1 <- vector(length = n.rep, mode = "numeric")
Label.2 <- rep(labels[2], n.rep)
Mean.2 <- vector(length = n.rep, mode = "numeric")
BF.2 <- vector(length = n.rep, mode = "numeric")
cat("done\n")
##################################################################
# (ii) Calculation of means and Bayes factors of these samples
##################################################################
cat("Calculating sample means and Bayes factors...\t")
for(i in 1 : n.rep)
{
	sample.1 <- rlogspline(n.sample, fit.1) # Sample from the resistant distribution function
	sample.2 <- rlogspline(n.sample, fit.2) # Sample from the susceptible distribution function
	BF.1[i] <- prod(dlogspline(sample.1, fit.1)) / prod(dlogspline(sample.1, fit.2)) # Bayes factor for resistant sample
	BF.2[i] <- prod(dlogspline(sample.2, fit.1)) / prod(dlogspline(sample.2, fit.2)) # Bayes factor for susceptible sample
	Mean.1[i] <- mean(sample.1) # Mean measurement of resistant sample
	Mean.2[i] <- mean(sample.2) # Mean measurement of susceptible sample		
}
cat("done\n")
# Assembly of matrix A containing means, Bayes factors, and labels. Score is maximum value minus mean value
Mean <- c(Mean.1, Mean.2)
BF <- c(BF.1, BF.2)
Label <- c(Label.1, Label.2)

A <- data.frame(Mean = Mean, BF = BF, Label = Label)

# Auxiliary variables
A$Label.Bool <- A$Label == labels[1] # Label must be Boolean for rocit. "resistant" is set to TRUE, "susceptible" is set to FALSE
A$Score <- x.max - A$Mean # In ROC analysis the predictor must point towards increasing probability of being resistant

##################################################################
# (iii) ROC analysis to find optimal cutoff values for mean and Bayes factor
##################################################################

cat("ROC analysis...\n")
predictors <- c("Score", "BF"); n.pred <- length(predictors)
# Vectors for results and various types of data
FPR <- c(); FPR.lo <- c(); FPR.up <-c(); TPR <- c(); TPR.lo <- c(); TPR.up <- c(); Predictor <- c()
AUC <- c(); SampleSize <- c(); CutOff <- c()
AUC.lo <- c(); AUC.up <- c()
#count <- 0 # counter for the predicors

for(count in 1 : n.pred) # number of predictors, here we have 2 predictors namely sample mean and Bayes factor
{
	predictor <- predictors[count]
	
	# cat("\n########################################################################\n")
	# cat("Using", predictor, "as predictor\n")
	# count <- count + 1
	Predictor <- c(Predictor, predictor)
				
	ROCit_obj <- rocit(score = A[, predictor], class = A$Label.Bool)
	plot.out <- plot(ROCit_obj) # rocit is primarily a plotting function
	# Pulling out the relevant pieces from the rocit object
	auc <- plot.out[[2]];			AUC[count] <- auc	
	fpr.opt <- plot.out[[6]][2];		FPR[count] <- fpr.opt			
	tpr.opt <- plot.out[[6]][3];		TPR[count] <- tpr.opt
	if(predictor == "Score") { cutoff <- x.max - plot.out[[6]][4] } # Back-transformation to average OD cutoff
	else { cutoff <- plot.out[[6]][4] }
	CutOff[count] <- cutoff		
}

# Setting up table with FPR, TPR, AUC at the optimal cutoff values for the predictors mean and Bayes factor 
cat("\n\nResult table\n")
Predictors2 <- c("Mean", "Bayes Factor")
Result <- data.frame(Predictor = Predictors2, CutOff, FPR, TPR)
print(Result)



################################################################################################################################################################################
#
# Classification of new data
# (i) using sample means
# (ii) using Bayes factor
# as classifier
#
################################################################################################################################################################################

# (i) Classification using sampe means cutoff
m <- D.test$Mean < Result$CutOff[1]
D.test$PredMean[m] <- "resistant"
D.test$PredMean[!m] <- "susceptible"

# (ii) Classification using Bayes factors
m <- D.test$BF > Result$CutOff[2]
D.test$PredBF[m] <- "resistant"
D.test$PredBF[!m] <- "susceptible"


cat("\n\nClassification statistics\n")

cat("Using mean as predictor\n")
print(table(D.test$PredMean))
cat("\n\nUsing Bayesian factors as predictor\n")
print(table(D.test$PredBF))

write.table(D.test, file = prediction.file, row.names = F, sep = "\t", quote = F)


